"""Example to diametrize single morphologie."""
import matplotlib.pyplot as plt
import numpy as np
from morphio.mut import Morphology
from neurom import NeuriteType
from neurom import load_morphology
from neurom import view

from diameter_synthesis.main import diametrize_single_neuron

if __name__ == "__main__":
    morph = Morphology("PATH_TO_MORPHOLOGY")
    smooth_morph = diametrize_single_neuron(morph)
    morph = load_morphology(morph)
    new_morph = load_morphology(smooth_morph)

    plt.figure()
    ax = plt.gca()
    view.plot_morph(
        morph.transform(lambda p: p - np.array([500, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
        neurite_type=NeuriteType.apical_dendrite,
    )
    view.plot_morph(
        new_morph.transform(lambda p: p + np.array([500, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
        neurite_type=NeuriteType.apical_dendrite,
    )
    view.plot_morph(
        morph.transform(lambda p: p - np.array([500, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
        neurite_type=NeuriteType.basal_dendrite,
    )
    view.plot_morph(
        new_morph.transform(lambda p: p + np.array([500, 0, 0])),
        ax=ax,
        plane="xy",
        soma_outline=False,
        realistic_diameters=False,
        neurite_type=NeuriteType.basal_dendrite,
    )
    plt.autoscale()
    plt.axis("equal")
    plt.savefig("compare_rediametetrized.pdf", bbox_inches="tight")



"""Tools to validate the input parameters."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json

from jsonschema import validate
from pkg_resources import resource_stream

MODEL_SCHEMA = json.load(resource_stream("diameter_synthesis", "schemas/model_params.json"))
CONFIG_SCHEMA = json.load(resource_stream("diameter_synthesis", "schemas/config.json"))


def validate_model_params(data):
    """Validate model parameter dictionary."""
    validate(data, MODEL_SCHEMA)


def validate_config(data):
    """Validate config dictionary."""
    validate(data, CONFIG_SCHEMA)



"""Construct diameter models from cells."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from functools import partial

from neurom import NeuriteType

import diameter_synthesis.morph_functions as morph_funcs
from diameter_synthesis.distribution_fitting import fit_distribution
from diameter_synthesis.exception import DiameterSynthesisError
from diameter_synthesis.simpler_diametrizer import build_simpler_model

L = logging.getLogger(__name__)


def _get_model_builder(config):
    """Get the diameter model builder.

    Args:
        config (dict): configuration dictionary.

    Returns:
        dict: dictionary of models.
    """
    all_models = {}
    for model in config["models"]:
        if model == "simpler":
            all_models[model] = build_simpler_model
        else:
            if model == "generic":
                distribution_types = {
                    "sibling_ratios": ["expon_rev", None],
                    "diameter_power_relation": ["exponnorm", None],
                    "terminal_diameters": ["exponnorm", None],
                    "trunk_diameters": ["exponnorm", None],
                    "trunk_diameters_method": "last",
                    "tapers": ["expon_rev", None],
                }
            elif model == "astrocyte":
                distribution_types = {
                    "sibling_ratios": ["expon_rev", None],
                    "diameter_power_relation": ["exponnorm", None],
                    "terminal_diameters": ["exponnorm", None],
                    "trunk_diameters_method": "mean",
                    "trunk_diameters": ["exponnorm", None],
                    "tapers": ["exponnorm", None],
                }
            else:
                raise DiameterSynthesisError(f"model not understood {model}")

            all_models[model] = partial(build_single_model, distribution_types)
    return all_models


def build(morphologies, config, with_data=False):
    """Build a model from a set of morphologies and a config file.

    Args:
        morphologies (dict): dictionary of morphologies.
        config (dict): configuration dictionary.
        with_data (bool): if set to True, the model data are also returned.

    Returns:
        dict: parameter of the models (and possibly the data extracted if `with_data==True`).
    """
    all_models = _get_model_builder(config)
    models_params, models_data = all_models[config["models"][0]](morphologies, config)
    if with_data:
        return models_params, models_data
    return models_params


def build_single_model(sampling_model, morphologies, config):
    """Build a single model.

    Args:
        sampling_model (dict): parameters for model building.
        morphologies (dict): dictionary of morphologies.
        config (dict): configuration dictionary.

    Returns:
        dict: parameter of the models data extracted.
    """
    all_data = extract_parameters(sampling_model, morphologies, config)
    all_models = fit_all_models(all_data, sampling_model, config)
    return all_models, all_data


def extract_parameters(sampling_model, morphologies, config):
    """Extract parameters from neurites to then use for model building.

    Args:
        sampling_model (dict): parameters for model building.
        morphologies (dict): dictionary of morphologies.
        config (dict): configuration dictionary.

    Returns:
        dict: data extracted.
    """
    all_data = {
        "sibling_ratios": {},
        "diameter_power_relation": {},
        "terminal_diameters": {},
        "trunk_diameters": {},
        "tapers": {},
    }

    for neurite_type in config["neurite_types"]:
        all_data["sibling_ratios"][neurite_type] = []
        all_data["diameter_power_relation"][neurite_type] = []
        all_data["terminal_diameters"][neurite_type] = []
        all_data["trunk_diameters"][neurite_type] = []
        all_data["tapers"][neurite_type] = []

        def _float(data):
            return [float(d) for d in data]

        for neuron in morphologies:
            for neurite in neuron.neurites:
                if neurite.type == getattr(NeuriteType, neurite_type):
                    all_data["sibling_ratios"][neurite_type] += _float(
                        morph_funcs.compute_sibling_ratios(
                            neurite, attribute_name=sampling_model["sibling_ratios"][1]
                        )
                    )
                    all_data["diameter_power_relation"][neurite_type] += _float(
                        morph_funcs.compute_diameter_power_relation(
                            neurite,
                            attribute_name=sampling_model["diameter_power_relation"][1],
                        )
                    )
                    all_data["terminal_diameters"][neurite_type] += _float(
                        morph_funcs.terminal_diameters(
                            neurite,
                            threshold=config["terminal_threshold"],
                            attribute_name=sampling_model["terminal_diameters"][1],
                        )
                    )
                    all_data["trunk_diameters"][neurite_type] += _float(
                        morph_funcs.trunk_diameter(
                            neurite,
                            attribute_name=sampling_model["trunk_diameters"][1],
                            method=sampling_model["trunk_diameters_method"],
                        )
                    )
                    all_data["tapers"][neurite_type] += _float(
                        morph_funcs.taper(
                            neurite,
                            params=config["taper"],
                            attribute_name=sampling_model["tapers"][1],
                        )
                    )
    return all_data


def fit_all_models(all_data, sampling_model, config):
    """Fit the parameters to get models.

    Args:
        all_data (dict): parameters to fit.
        sampling_model (dict): parameters for model building.
        config (dict): configuration dictionary.

    Returns:
        dict: models.
    """
    all_models = {
        "sibling_ratios": {},
        "diameter_power_relation": {},
        "terminal_diameters": {},
        "trunk_diameters": {},
        "tapers": {},
    }

    extra_params = config.copy()
    for neurite_type in config["neurite_types"]:

        extra_params["neurite_type"] = neurite_type
        extra_params["name"] = "sibling_ratios"
        all_models["sibling_ratios"][neurite_type] = fit_model(
            sampling_model["sibling_ratios"],
            all_data["sibling_ratios"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "diameter_power_relation"
        all_models["diameter_power_relation"][neurite_type] = fit_model(
            sampling_model["diameter_power_relation"],
            all_data["diameter_power_relation"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "terminal_diameter"
        all_models["terminal_diameters"][neurite_type] = fit_model(
            sampling_model["terminal_diameters"],
            all_data["terminal_diameters"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "trunk_diameters"
        all_models["trunk_diameters"][neurite_type] = fit_model(
            sampling_model["trunk_diameters"],
            all_data["trunk_diameters"][neurite_type],
            extra_params,
        )

        extra_params["name"] = "tapers"
        all_models["tapers"][neurite_type] = fit_model(
            sampling_model["tapers"], all_data["tapers"][neurite_type], extra_params
        )

    return all_models


def fit_model(sampling_model, data, extra_params):
    """Fit a single parameter.

    Args:
        sampling_model (dict): model parameters.
        data (array): parameter to fit.
        extra_params (dict): additional parameters.

    Returns:
        dict: fit parameters.
    """
    if len(data) == 0:
        return {}
    return {
        "distribution": sampling_model[0],
        "sequential": sampling_model[1],
        "params": fit_distribution(
            data,
            sampling_model[0],
            attribute_name=sampling_model[1],
            extra_params=extra_params,
        ),
    }



"""Test the build_diameters module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import morphio.mut
import numpy as np
import pytest
from morphio import PointLevel
from morphio import SectionType

from diameter_synthesis import build_diameters
from diameter_synthesis.exception import DiameterSynthesisError

from .testing_tools import _compare_diameters

NEURITE_TYPES = ["basal_dendrite", "apical_dendrite"]


def test_build_simple(config, model_params, small_morph, test_data_path):
    """Test the main build function to diametrize a neuron."""
    build_diameters.build(small_morph, NEURITE_TYPES, model_params, config)
    expected = morphio.mut.Morphology(test_data_path / "small_morph_diametrized.asc")
    _compare_diameters(expected, small_morph)


def test_build_simple_with_apical_sections(config, model_params, small_morph, test_data_path):
    """Test the main build function to diametrize a neuron with a known apical section."""
    model_params["apical_point_sec_ids"] = [6]

    build_diameters.build(small_morph, NEURITE_TYPES, model_params, config)
    expected = morphio.mut.Morphology(test_data_path / "small_morph_apical_diametrized.asc")
    _compare_diameters(expected, small_morph)


def test_build_simple_with_several_apical_sections(
    config, model_params, small_morph, test_data_path
):
    """Test the main build function to diametrize a neuron with a known apical section."""
    # Duplicate the apical
    for section in small_morph.sections[4].iter():
        if section.id != 4:
            append_fun = small_morph.sections[section.parent.id + 13].append_section
        else:
            append_fun = small_morph.append_root_section
        pts = section.points.copy()
        pts[:, 0] += 1000
        append_fun(PointLevel(pts, section.diameters.copy()), SectionType(3))

    model_params["apical_point_sec_ids"] = [6, 19]

    build_diameters.build(small_morph, NEURITE_TYPES, model_params, config)

    expected = morphio.mut.Morphology(test_data_path / "small_morph_several_apical_diametrized.asc")
    _compare_diameters(expected, small_morph)


def test_build(config, model_params, neuron, neuron_diametrized):
    """Test the main build function to diametrize a neuron."""
    build_diameters.build(neuron, NEURITE_TYPES, model_params, config)
    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_seed(config, model_params, neuron, neuron_diametrized):
    """Test main build without seed in config."""
    rng = np.random.default_rng(config.pop("seed"))

    build_diameters.build(neuron, NEURITE_TYPES, model_params, config, rng)
    _compare_diameters(neuron_diametrized, neuron)


def test_build_multiple_models_warning(config, model_params, neuron, neuron_diametrized, caplog):
    """Test main build with multiple models in config."""
    config["models"].append("SKIPPED MODEL")

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, NEURITE_TYPES, model_params, config)

    module, level, entry = caplog.record_tuples[0]
    assert module == "diameter_synthesis.build_diameters"
    assert level == 30
    assert entry == "Several models provided, we will only use the first"

    _compare_diameters(neuron_diametrized, neuron)


def test_build_one_sample(test_data_path, config, model_params, neuron):
    """Test main build with 1 sample in config."""
    config["n_samples"] = 1

    build_diameters.build(neuron, NEURITE_TYPES, model_params, config)

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_1_sample.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_no_sequential(test_data_path, config, model_params, neuron):
    """Test main build with no sequential in config."""
    config["models"] = ["astrocyte"]
    config["n_samples"] = 1
    model_params["sibling_ratios"]["apical_dendrite"]["sequential"] = None

    build_diameters.build(neuron, NEURITE_TYPES, model_params, config)

    neuron_diametrized = morphio.mut.Morphology(
        test_data_path / "C030796A-P3_lite_diametrized_astrocyte.h5"
    )

    _compare_diameters(neuron_diametrized, neuron)


def test_build_small_n_tries_warning(config, model_params, neuron, caplog):
    """Test main build with trunk_max_tries = 1."""
    config["trunk_max_tries"] = 1

    caplog.clear()
    caplog.set_level(logging.WARNING)
    build_diameters.build(neuron, NEURITE_TYPES, model_params, config)
    assert len(caplog.record_tuples) == 5
    neurite_types = [
        "apical_dendrite",
        "apical_dendrite",
        "apical_dendrite",
        "apical_dendrite",
        "apical_dendrite",
    ]
    for i in zip(neurite_types, caplog.record_tuples):
        neurite_type, (module, level, entry) = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == f"max tries attained with {neurite_type}"


def test_build_small_trunk_diam_warning(config, model_params, neuron, caplog):
    """Test trunk diam warning in main build."""
    TRUNK_FRAC_DECREASE = build_diameters.TRUNK_FRAC_DECREASE
    try:
        build_diameters.TRUNK_FRAC_DECREASE = 999
        caplog.clear()
        caplog.set_level(logging.WARNING)
        build_diameters.build(neuron, NEURITE_TYPES, model_params, config)
    finally:
        build_diameters.TRUNK_FRAC_DECREASE = TRUNK_FRAC_DECREASE

    assert len(caplog.record_tuples) == 1
    for i in caplog.record_tuples:
        module, level, entry = i
        assert module == "diameter_synthesis.build_diameters"
        assert level == 30
        assert entry == "sampled trunk diameter < 0.01, so use 1 instead"


def test_select_model():
    """Test the _select_model function."""
    # pylint: disable=protected-access
    # pylint: disable=no-member
    f_generic = build_diameters._select_model("generic")
    assert f_generic.args == (
        {
            "mode_sibling": "threshold",
            "mode_diameter_power_relation": "threshold",
            "with_asymmetry": True,
            "reduction_factor_max": 1.0,
        },
    )

    f_astrocyte = build_diameters._select_model("astrocyte")
    assert f_astrocyte.args == (
        {
            "mode_sibling": "generic",
            "mode_diameter_power_relation": "generic",
            "with_asymmetry": True,
            "reduction_factor_max": 3.0,
        },
    )

    with pytest.raises(DiameterSynthesisError):
        build_diameters._select_model("UNKNOWN")


def test_sample_sibling_ratio(model_params):
    """Test the _sample_sibling_ratio function."""
    # pylint: disable=protected-access
    d_generic = build_diameters._sample_sibling_ratio(
        model_params, "basal_dendrite", False, mode="generic"
    )
    assert d_generic == pytest.approx(0.6925189949520985)

    d_threshold = build_diameters._sample_sibling_ratio(
        model_params, "basal_dendrite", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(0.5222098881400881)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_sibling_ratio(model_params, "basal_dendrite", False, mode="UNKNOWN")


def test_sample_diameter_power_relation(model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    d_generic = build_diameters._sample_diameter_power_relation(
        model_params, "basal_dendrite", False, mode="generic"
    )
    assert d_generic == pytest.approx(5.281970058326886)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params, "basal_dendrite", False, mode="threshold"
    )
    assert d_threshold == pytest.approx(5.915270977501953)

    d_threshold = build_diameters._sample_diameter_power_relation(
        model_params, "basal_dendrite", False, mode="exact"
    )
    assert d_threshold == pytest.approx(1)

    with pytest.raises(DiameterSynthesisError):
        build_diameters._sample_diameter_power_relation(
            model_params, "basal_dendrite", False, mode="UNKNOWN"
        )


def test_sample_daughter_diameters(neuron, model_params):
    """Test the _sample_diameter_power_relation function."""
    # pylint: disable=protected-access
    param_tree = {
        "mode_sibling": "threshold",
        "mode_diameter_power_relation": "threshold",
        "with_asymmetry": True,
        "reduction_factor_max": 1.0,
        "neurite_type": "basal_dendrite",
        "asymmetry_threshold": 1.0,
        "trunk_diam": 2.131444808766456,
        "tot_length": 1085.1178037524223,
        "terminal_diam": 0.35602967635703164,
        "major_sections": [],
    }
    d = build_diameters._sample_daughter_diameters(neuron.sections[0], model_params, param_tree)
    assert d == pytest.approx([0.9856948, 0.68261237])

    param_tree["with_asymmetry"] = False
    d_no_asymmetry = build_diameters._sample_daughter_diameters(
        neuron.sections[0], model_params, param_tree
    )
    assert d_no_asymmetry == pytest.approx([1.23559368, 0.77282112])


def test_diametrize_axon(small_morph):
    """Test the axon diametrizer."""
    build_diameters.diametrize_axon(small_morph)

    expected = morphio.mut.Morphology(small_morph)
    diameters = {
        0: [0.2, 0.2, 0.2],  # basal
        1: [0.2, 0.2, 0.2],  # basal
        2: [0.2, 0.2, 0.2],  # basal
        3: [0.2, 0.2, 0.2],  # basal
        4: [0.2, 0.2, 0.2],  # apical
        5: [0.2, 0.2, 0.2],  # apical
        6: [0.2, 0.2, 0.2],  # apical
        7: [0.2, 0.2, 0.2],  # apical
        8: [0.2, 0.2, 0.2],  # apical
        9: [0.2, 0.2, 0.2],  # apical
        10: [0.2, 0.2, 0.2],  # apical
        11: [0.2, 0.2, 0.2],  # apical
        12: [0.2, 0.2, 0.2],  # apical
        13: [0.2, 0.2, 0.9],  # axon
        14: [0.2, 0.2, 0.9],  # axon
        15: [0.376628, 0.326628, 0.276628],  # basal
        16: [0.753256, 0.703256, 0.653256],  # basal
    }
    for section_id, section in expected.sections.items():
        section.diameters = diameters[section_id]

    _compare_diameters(expected, small_morph, rtol=1e-6)



"""Test the build_models module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import morphio.mut

from diameter_synthesis import build_diameters
from diameter_synthesis import build_models

from .testing_tools import _compare_diameters
from .testing_tools import compare_dicts


def test_build_model(single_pop, simpler_model_params, simpler_model_data):
    """Test the build function."""
    # Test with generic model
    config = {
        "models": ["simpler"],
        "neurite_types": ["basal_dendrite", "apical_dendrite"],
    }
    res = build_models.build(single_pop, config, with_data=True)
    res_models_params = build_models.build(single_pop, config, with_data=False)

    assert len(res) == 2
    assert compare_dicts(res, [simpler_model_params, simpler_model_data], precision=3)
    assert compare_dicts(res_models_params, simpler_model_params, precision=3)


def test_build_diameters(simpler_config, simpler_model_params, small_morph, test_data_path):
    """Test the main build function to diametrize a neuron."""
    neurite_types = ["basal_dendrite", "apical_dendrite"]

    build_diameters.build(small_morph, neurite_types, simpler_model_params, simpler_config)
    expected = morphio.mut.Morphology(test_data_path / "simpler_morph_diametrized.asc")
    _compare_diameters(expected, small_morph)



"""Test the plotting module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import neurom as nm
import pytest
from diff_pdf_visually import pdf_similar
from matplotlib import pyplot as plt

from diameter_synthesis import plotting
from diameter_synthesis.exception import DiameterSynthesisError

NEURITE_TYPES = ["basal_dendrite", "apical_dendrite"]


def test_plot_diameter_diff(neuron_diametrized_path, tmpdir, expected_images_path):
    """Test the plot_diameter_diff function."""
    # Test with existing directory and with positive changes

    # Load neuron with NeuroM
    new_neuron = nm.load_morphology(neuron_diametrized_path)

    # Update the neuron
    for sec in new_neuron.sections:
        sec.points[:, nm.COLS.R] *= 20

    # Plot the figure
    plotting.plot_diameter_diff(
        neuron_diametrized_path, new_neuron, NEURITE_TYPES, tmpdir, ext=".pdf"
    )

    # Check the figure
    assert pdf_similar(
        str(expected_images_path / "test_plot_diameter_diff_mult.pdf"),
        str(tmpdir / neuron_diametrized_path.with_suffix(".pdf").name),
    )

    # Test with new directory and with negative changes

    # Load neuron with NeuroM
    new_neuron = nm.load_morphology(neuron_diametrized_path)

    # Update the neuron
    for sec in new_neuron.sections:
        sec.points[:, nm.COLS.R] /= 20

    # Plot the figure
    plotting.plot_diameter_diff(
        neuron_diametrized_path, new_neuron, NEURITE_TYPES, tmpdir / "new_dir", ext=".pdf"
    )

    # Check the figure
    assert pdf_similar(
        str(expected_images_path / "test_plot_diameter_diff_div.pdf"),
        str(tmpdir / "new_dir" / neuron_diametrized_path.with_suffix(".pdf").name),
    )


def test_plot_distribution_fit(model_params, model_data, tmpdir, expected_images_path):
    """Test the plot_distribution_fit function."""
    # Plot the figure
    plotting.plot_distribution_fit(
        model_data["diameter_power_relation"],
        model_params["diameter_power_relation"],
        NEURITE_TYPES,
        fig_name=tmpdir / "test_plot_distribution_fit",
        ext=".pdf",
    )

    # Check the figures
    assert pdf_similar(
        str(expected_images_path / "test_plot_distribution_fit.pdf"),
        str(tmpdir / "test_plot_distribution_fit.pdf"),
    )


def test_plot_cumulative_distribution(
    single_pop, single_pop_diametrized, tmpdir, expected_images_path
):
    """Test the plot_cumulative_distribution function."""
    # Plot the figure
    plotting.plot_cumulative_distribution(
        single_pop,
        single_pop_diametrized,
        "segment_radial_distances",
        "segment_volumes",
        NEURITE_TYPES,
    )
    plt.savefig(tmpdir / "test_plot_cumulative_distribution.pdf", bbox_inches="tight")
    plt.close()

    # Check the figures
    assert pdf_similar(
        str(expected_images_path / "test_plot_cumulative_distribution.pdf"),
        str(tmpdir / "test_plot_cumulative_distribution.pdf"),
    )


def test_make_cumulative_figures(single_pop, single_pop_diametrized, tmpdir, expected_images_path):
    """Test the make_cumulative_figures function."""
    # Plot the figures
    plotting.make_cumulative_figures(
        single_pop,
        single_pop_diametrized,
        "segment_radial_distances",
        "segment_volumes",
        NEURITE_TYPES,
        tmpdir,
        individual=True,
        figname_prefix="with_individual_",
        ext=".pdf",
    )

    # Check the figures
    images_path = expected_images_path / "test_make_cumulative_figures"
    assert pdf_similar(
        str(images_path / "cumulative_segment_radial_distances_volumes.pdf"),
        str(tmpdir / "with_individual_cumulative_segment_radial_distances_volumes.pdf"),
    )
    assert pdf_similar(
        str(
            images_path
            / "cumulative_segment_radial_distances_volumes_individual"
            / "0_cumulative_segment_radial_distances_volumes.pdf"
        ),
        str(
            tmpdir
            / "with_individual_cumulative_segment_radial_distances_volumes_individual"
            / (
                "0_with_individual_cumulative_segment_radial_distances_volumes_C030796A-P3_lite"
                ".h5.pdf"
            )
        ),
    )


def test_cumulative_analysis(
    single_pop_neurondb,
    tmpdir,
    expected_images_path,
    single_pop_data_dir,
    single_pop_diametrized_data_dir,
):
    """Test the cumulative_analysis function."""
    # Plot the figures
    plotting.cumulative_analysis(
        single_pop_data_dir,
        single_pop_diametrized_data_dir,
        tmpdir / "analysis",
        True,
        single_pop_data_dir / "neurondb.dat",
        NEURITE_TYPES,
        ext=".pdf",
    )

    # Check the figures
    images_path = expected_images_path / "test_cumulative_analysis" / "analysis"
    assert pdf_similar(
        str(images_path / "L5_TPC_A_cumulative_section_path_distances_areas.pdf"),
        str(tmpdir / "analysis" / "L5_TPC_A_cumulative_section_path_distances_areas.pdf"),
    )
    assert pdf_similar(
        str(images_path / "L5_TPC_A_cumulative_section_path_distances_volumes.pdf"),
        str(tmpdir / "analysis" / "L5_TPC_A_cumulative_section_path_distances_volumes.pdf"),
    )
    assert pdf_similar(
        str(
            images_path
            / "L5_TPC_A_cumulative_section_path_distances_areas_individual"
            / "0_L5_TPC_A_cumulative_section_path_distances_areas_C030796A-P3_lite.pdf"
        ),
        str(
            tmpdir
            / "analysis"
            / "L5_TPC_A_cumulative_section_path_distances_areas_individual"
            / "0_L5_TPC_A_cumulative_section_path_distances_areas_C030796A-P3_lite.h5.pdf"
        ),
    )
    assert pdf_similar(
        str(
            images_path
            / "L5_TPC_A_cumulative_section_path_distances_volumes_individual"
            / "0_L5_TPC_A_cumulative_section_path_distances_volumes_C030796A-P3_lite.pdf"
        ),
        str(
            tmpdir
            / "analysis"
            / "L5_TPC_A_cumulative_section_path_distances_volumes_individual"
            / "0_L5_TPC_A_cumulative_section_path_distances_volumes_C030796A-P3_lite.h5.pdf"
        ),
    )

    # Test with inconsistent neurondb
    single_pop_neurondb.loc[0, "morphology"] = "UNKNOWN_MORPHOLOGY"
    single_pop_neurondb.to_csv(
        single_pop_data_dir / "neurondb.dat", sep=" ", header=False, index=False
    )
    with pytest.raises(DiameterSynthesisError):
        plotting.cumulative_analysis(
            single_pop_data_dir,
            single_pop_diametrized_data_dir,
            tmpdir / "analysis",
            True,
            single_pop_data_dir / "neurondb.dat",
            NEURITE_TYPES,
            ext=".pdf",
        )


def test_violin_analysis(
    tmpdir, expected_images_path, single_pop_data_dir, single_pop_diametrized_data_dir
):
    """Test the violin_analysis function."""
    # Plot the figures
    plotting.violin_analysis(
        single_pop_data_dir,
        single_pop_diametrized_data_dir,
        tmpdir / "analysis",
        single_pop_data_dir / "neurondb.dat",
    )

    # Check the figures
    assert pdf_similar(
        str(expected_images_path / "test_violin_analysis.pdf"),
        str(tmpdir / "analysis" / "morphometrics.pdf"),
    )



"""Build neurite diameters from a pre-generated model."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging
from collections import deque
from copy import copy
from copy import deepcopy
from functools import partial

import numpy as np
from morphio import IterType
from morphio import SectionType
from numpy.polynomial import polynomial

from diameter_synthesis import morph_functions
from diameter_synthesis import utils
from diameter_synthesis.distribution_fitting import sample_distribution
from diameter_synthesis.exception import DiameterSynthesisError
from diameter_synthesis.simpler_diametrizer import simpler_diametrizer

TRUNK_FRAC_DECREASE = 0.1
N_TRIES_BEFORE_REDUC = 5
L = logging.getLogger(__name__)


def _reset_caches():
    """Reset the cached functions."""
    morph_functions.sec_length.cache_clear()
    morph_functions.partition_asymmetry_length.cache_clear()
    morph_functions.lengths_from_origin.cache_clear()
    morph_functions.n_children_downstream.cache_clear()


def _get_neurites(neuron, neurite_type):
    """Get a list of neurites to diametrize.

    Args:
        neuron (morphio.mut.Morphology): a neuron.
        neurite_type (morphio.SectionType): the neurite type to consider.

    Returns:
        list: list of neurites to consider.
    """
    return [
        list(neurite.iter()) for neurite in neuron.root_sections if neurite.type == neurite_type
    ]


def _sample_sibling_ratio(
    params, neurite_type, apply_asymmetry=False, mode="generic", rng=np.random
):
    """Sample a sibling ratio from distribution.

    Args:
        params (dict): model parameters.
        neurite_type (str): the neurite type to consider.
        apply_asymmetry (bool): asymmetry of current branching point.
        mode (str): to use or not the `asymmetry_threshold`.
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        float: sibling ratio.
    """
    if mode == "generic":
        return sample_distribution(params["sibling_ratios"][neurite_type], rng=rng)
    if mode == "threshold":
        if apply_asymmetry:
            return 0.0
        return sample_distribution(params["sibling_ratios"][neurite_type], rng=rng)
    # This case should never happen since the mode is already checked in `_select_model`
    raise DiameterSynthesisError(f"mode not understood {mode}")


def _sample_diameter_power_relation(
    params, neurite_type, apply_asymmetry=False, mode="generic", rng=np.random
):
    """Sample a diameter power relation from distribution.

    Args:
        params (dict): model parameters.
        neurite_type (str): the neurite type to consider.
        apply_asymmetry (bool): asymmetry of current branching point.
        mode (str): to use or not the `asymmetry_threshold`.
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        float: diameter power relation.
    """
    if mode == "generic":
        return sample_distribution(params["diameter_power_relation"][neurite_type], rng=rng)
    if mode == "threshold":
        if apply_asymmetry:
            return 1.0
        return sample_distribution(params["diameter_power_relation"][neurite_type], rng=rng)
    if mode == "exact":
        # This case should never happen since this mode is not known by `_select_model`
        return 1.0
    # This case should never happen since the mode is already checked in `_select_model`
    raise DiameterSynthesisError(f"mode not understood {mode}")


def _sample_trunk_diameter(params, neurite_type, rng=np.random):
    """Sample a trunk diameter from distribution.

    Args:
        params (dict): model parameters.
        neurite_type (str): the neurite type to consider.
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        float: trunk diameter
    """
    return sample_distribution(params["trunk_diameters"][neurite_type], rng=rng)


def _sample_terminal_diameter(params, neurite_type, rng=np.random):
    """Sample a terminal diameter.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        float: terminal diameter
    """
    return sample_distribution(params["terminal_diameters"][neurite_type], rng=rng)


def _sample_taper(params, neurite_type, rng=np.random):
    """Sample a taper rate from distributions.

    Args:
        params (dict): model parameters
        neurite_type (str): the neurite type to consider
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        float: taper rate
    """
    return sample_distribution(params["tapers"][neurite_type], rng=rng)


def _sample_daughter_diameters(section, params, params_tree, rng=np.random):
    """Compute the daughter diameters of the current section.

    Args:
        section (morphio.Section): section to consider.
        params (dict): model parameters.
        params_tree (dict): specific parameters of the current tree.
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
       list: list of daughter diameters.
    """
    # pylint: disable=too-many-locals
    major_sections = params_tree["major_sections"]

    apply_asymmetry = False
    if params_tree.get("with_asymmetry", False):
        child_not_in_major = [child.id not in major_sections for child in section.children]
        # if children are not asymmetrical, don't apply asymmetry
        if False in child_not_in_major:
            child_sort = np.argsort(child_not_in_major)
            apply_asymmetry = True

    reduction_factor = params_tree["reduction_factor_max"] + 1.0
    # try until we get a reduction of diameter in the branching
    while reduction_factor > params_tree["reduction_factor_max"]:

        sibling_ratio = _sample_sibling_ratio(
            params,
            params_tree["neurite_type"],
            apply_asymmetry=apply_asymmetry,
            mode=params_tree["mode_sibling"],
            rng=rng,
        )

        diameter_power_relation = _sample_diameter_power_relation(
            params,
            params_tree["neurite_type"],
            apply_asymmetry=apply_asymmetry,
            mode=params_tree["mode_diameter_power_relation"],
            rng=rng,
        )

        reduction_factor = morph_functions.diameter_power_relation_factor(
            diameter_power_relation, sibling_ratio
        )

    diam_0 = section.diameters[-1]
    terminal_diam = min(diam_0, params_tree["terminal_diam"])

    diam_1 = reduction_factor * diam_0
    diam_2 = sibling_ratio * diam_1

    diam_1 = max(diam_1, terminal_diam)
    diam_2 = max(diam_2, terminal_diam)

    diams = np.array([diam_1] + (len(section.children) - 1) * [diam_2])

    if apply_asymmetry:
        sorted_diams = np.empty_like(diams)
        sorted_diams[child_sort] = diams
        return sorted_diams

    # At the moment we don't have enough information to do better than a random choice in this case
    rng.shuffle(diams)
    return diams


def _diametrize_section(section, initial_diam, taper, min_diam=0.07, max_diam=10.0):
    """Diameterize a section.

    Args:
        section (morphio.Section): current section.
        initial_diam (float): initial diameter.
        taper (float): taper rate.
        min_diam (float): minimum diameter.
        max_diam (float): maximum diameter.
    """
    diams = polynomial.polyval(morph_functions.lengths_from_origin(section), [initial_diam, taper])
    section.diameters = np.clip(diams, min_diam, max_diam)


def _diametrize_tree(neurite, params, params_tree, rng=np.random):
    """Diametrize a tree, or neurite.

    Args:
        neurite (morphio neurite): current neurite.
        params (dict): model parameters.
        params_tree (dict): specific parameters of the current tree.
        rng (numpy.random.Generator): the random number generator to use.

    Returns:
        bool: `True` is all terminal diameters are small enough, `False` otherwise.
    """
    params_tree["tot_length"] = morph_functions.get_total_length(neurite)
    max_diam = params["terminal_diameters"][params_tree["neurite_type"]]["params"]["max"]
    wrong_tips = False
    active = deque([neurite[0]])
    while active:
        section = active.popleft()

        if section.is_root:
            init_diam = params_tree["trunk_diam"]
        else:
            init_diam = section.diameters[0]

        taper = _sample_taper(params, params_tree["neurite_type"], rng=rng)

        params_tree["terminal_diam"] = min(
            init_diam, _sample_terminal_diameter(params, params_tree["neurite_type"], rng=rng)
        )

        _diametrize_section(
            section,
            init_diam,
            taper=taper,
            min_diam=params_tree["terminal_diam"],
            max_diam=params_tree["trunk_diam"],
        )

        if len(section.children) > 0:
            diams = _sample_daughter_diameters(section, params, params_tree, rng=rng)

            for i, child in enumerate(section.children):
                utils.redefine_diameter_section(child, 0, diams[i])
                active.append(child)

        # if we are at a tip, check if tip diameters are small enough
        elif section.diameters[-1] > max_diam:
            wrong_tips = True

    return wrong_tips


def _diametrize_neuron(params_tree, neuron, params, neurite_types, config, rng=np.random):
    """Diametrize a neuron.

    Args:
        params_tree (dict): specific parameters of the current tree.
        neuron (morphio.mut.Morphology): neuron to diametrize.
        params (dict): model parameters.
        neurite_types (str or morphio.SectionType): the neurite type to consider.
        config (dict): general configuration parameters.
        rng (numpy.random.Generator): the random number generator to use.
    """
    # pylint: disable=too-many-locals, too-many-branches
    major_sections = set()
    if params_tree["with_asymmetry"]:
        # Get sections on the major branch
        for apical_section in params.get("apical_point_sec_ids", []):
            for sec in neuron.sections[apical_section].iter(IterType.upstream):
                major_sections.add(sec.id)
    params_tree["major_sections"] = major_sections

    for neurite_type in neurite_types:
        if isinstance(neurite_type, str):
            morphio_neurite_type = getattr(SectionType, neurite_type)
        else:
            morphio_neurite_type = neurite_type
            neurite_type = neurite_type.name

        params_tree["neurite_type"] = neurite_type

        for neurite in _get_neurites(neuron, morphio_neurite_type):
            wrong_tips = True
            n_tries = 0
            trunk_diam_frac = 1.0
            n_tries_step = 1

            _params = deepcopy(params) if "taper_increase" in params_tree else params
            while wrong_tips:
                if "taper_increase" in params_tree:
                    _params["tapers"][neurite_type]["params"]["scale"] *= params_tree[
                        "taper_increase"
                    ]
                    params_tree["trunk_diam"] = neurite[0].diameters[0]
                else:
                    trunk_diam = trunk_diam_frac * _sample_trunk_diameter(
                        _params, neurite_type, rng=rng
                    )
                    if trunk_diam < 0.01:
                        trunk_diam = 1.0
                        L.warning("sampled trunk diameter < 0.01, so use 1 instead")
                    params_tree["trunk_diam"] = trunk_diam

                wrong_tips = _diametrize_tree(neurite, _params, params_tree, rng=rng)

                # if we can't get a good model, reduce the trunk diameter progressively
                if n_tries > N_TRIES_BEFORE_REDUC * n_tries_step:
                    trunk_diam_frac -= TRUNK_FRAC_DECREASE
                    n_tries_step += 1
                if n_tries > config.get("trunk_max_tries", 100):
                    L.warning("max tries attained with %s", neurite_type)
                    wrong_tips = False
                n_tries += 1
    return neuron


def _select_model(model):
    """Select a diametrized model to use.

    Args:
        model (str): model name.

    Returns:
        function: diametrizer with specific `params_tree`.
    """
    if model == "simpler":

        return simpler_diametrizer

    if model == "generic":
        params_tree = {}
        params_tree["mode_sibling"] = "threshold"
        params_tree["mode_diameter_power_relation"] = "threshold"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 1.0
    elif model == "astrocyte":
        params_tree = {}
        params_tree["mode_sibling"] = "generic"
        params_tree["mode_diameter_power_relation"] = "generic"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 3.0
    elif model == "neurite_based":
        params_tree = {}
        params_tree["mode_sibling"] = "threshold"
        params_tree["mode_diameter_power_relation"] = "threshold"
        params_tree["with_asymmetry"] = True
        params_tree["reduction_factor_max"] = 1.0
        params_tree["taper_increase"] = 1.02

    else:
        raise DiameterSynthesisError(f"Unknown diameter model: {model}")

    return partial(_diametrize_neuron, params_tree)


def build(neuron, neurite_types, model_params, diam_params, random_generator=np.random):
    """Builder function for generating diameters of a neuron from the a diameter models.

    Args:
        neuron (morphio.mut.Morphology): neuron to diametrize.
        neurite_types (list[str] or str or list[SectionType] or SectionType): the neurite type(s)
            to consider.
        model_params (dict): model parameters.
        diam_params (dict): general configuration parameters.
        random_generator (numpy.random.Generator): the random number generator to use.
    """
    if isinstance(neurite_types, (str, SectionType)):
        neurite_types = [neurite_types]

    if "seed" in diam_params:
        random_generator = np.random.default_rng(diam_params["seed"])

    _reset_caches()

    if len(diam_params["models"]) > 1:
        L.warning("Several models provided, we will only use the first")
    diameter_generator = _select_model(diam_params["models"][0])

    neuron = diameter_generator(
        neuron, model_params, neurite_types, diam_params, rng=random_generator
    )
    n_samples = diam_params.get("n_samples", 1)
    if n_samples > 1:
        diameters = utils.get_all_diameters(neuron)
        for _ in range(n_samples - 1):
            neuron = diameter_generator(
                neuron, model_params, neurite_types, diam_params, rng=random_generator
            )
            for i, new_diams in enumerate(utils.get_all_diameters(neuron)):
                diameters[i] += new_diams
        for i, _ in enumerate(diameters):
            diameters[i] /= n_samples
        utils.set_all_diameters(neuron, diameters)


def _save_first_diams(morphology, length):
    """Save diameters in a dict up to length."""
    diams = {}
    for root_section in morphology.root_sections:
        if root_section.type == SectionType.axon:
            dist = 0
            prev_point = root_section.points[0]
            for section in root_section.iter():
                _diams = copy(section.diameters)
                for point in section.points:
                    dist += np.linalg.norm(point - prev_point)
                    prev_point = copy(point)
                    if dist >= length:
                        break
                diams[section.id] = _diams
                if dist >= length:
                    break

    return diams


def _set_first_diams(morphology, diams, length):
    """Set diameters from a dict up to length."""
    for root_section in morphology.root_sections:
        if root_section.type == SectionType.axon:
            dist = 0
            prev_point = root_section.points[0]
            for section in root_section.iter():
                current_diams = copy(section.diameters)
                old_diams = diams[section.id]
                for i, point in enumerate(section.points):
                    dist += np.linalg.norm(point - prev_point)
                    prev_point = copy(point)
                    current_diams[i] = old_diams[i]
                    if dist >= length:
                        break
                section.diameters = current_diams
                if dist >= length:
                    break


def diametrize_axon(
    morphology,
    main_diameter=1.0,
    collateral_diameter=0.1,
    main_taper=-0.0005,
    axon_point_isec=None,
    ais_length=60,
    rng=np.random,
):
    """Diametrize axon in place without learning from reconstructed axons.

    The main axon branch (from soma to axon point) will have a tapered diameter, and the
    colaterals a constant diameter.

    If an axon point is not provided, and main_diameter > collateral_diameter, the diameters will
    decrease with taper and bifurcations, with hardcoded parameters sibling_ratio = 0.5 and
    diameter_power_relation = 0.5.

    Args:
        morphology (morphio.mut.Morphology): morphology to diametrize.
        main_diameter (float): diameter of main axon branch (from soma to `axon_point_isec`).
        collateral_diameter (float): diameter of collateral branches.
        main_taper (float): taper rate of main branch (set to 0 for no taper, should be negative).
        axon_point_isec (int): morphio section id of axon point (see :mod:`morph_tool.axon_point`
            module).
        ais_length (float): length of ais for which we keep original diameters.
        rng (numpy.random.Generator): the random number generator to use.
    """
    model_params = {
        "trunk_diameters": {
            "axon": {
                "distribution": "constant",
                "params": {"value": main_diameter},
            }
        },
        "terminal_diameters": {
            "axon": {
                "distribution": "constant",
                "params": {"value": collateral_diameter, "max": main_diameter},
            }
        },
        "tapers": {
            "axon": {
                "distribution": "constant",
                "params": {"value": main_taper},
            }
        },
        "sibling_ratios": {
            "axon": {
                "distribution": "constant",
                "params": {"value": 0.5},
            }
        },
        "diameter_power_relation": {
            "axon": {
                "distribution": "constant",
                "params": {"value": 5.0},
            }
        },
    }
    if axon_point_isec:
        model_params["apical_point_sec_ids"] = [axon_point_isec]

    config = {"models": ["generic"], "trunk_max_tries": 1, "n_samples": 1}

    diams = _save_first_diams(morphology, ais_length)
    build(morphology, "axon", model_params, diam_params=config, random_generator=rng)
    _set_first_diams(morphology, diams, ais_length)



"""Simpler diametrizer."""
from functools import partial

import matplotlib.pyplot as plt
import numpy as np
from neurom import NeuriteType
from neurom import iter_sections
from neurom.core.morphology import Morphology
from neurom.core.morphology import Section
from numpy.polynomial import Polynomial
from scipy.stats import pearsonr


def section_path_length(section, cache):
    """Path length from section to root."""
    if section.id not in cache:
        cache[section.id] = sum(s.length for s in section.iupstream())
    return cache[section.id]


def _map_sections(fun, neurite, iterator_type=Section.ipreorder):
    """Map `fun` to all the sections."""
    return list(map(fun, iterator_type(neurite.root_node)))


def terminal_path_lengths(neurite, cache):
    """Get the path lengths to each terminal point."""
    return _map_sections(partial(section_path_length, cache=cache), neurite, Section.ileaf)


def build_simpler_model(morphologies, config, fit_orders=None):
    """Build diameter model."""
    neurite_types = config.get("neurite_types")
    if neurite_types is None:
        neurite_types = ["basal_dendrite", "apical_dendrite"]
    if fit_orders is None:
        fit_orders = {"basal_dendrite": 1, "apical_dendrite": 2, "axon": 1}

    coeffs = {}
    residues = {}
    all_diams = {n_type: [] for n_type in neurite_types}
    all_lengths = {n_type: [] for n_type in neurite_types}

    for neurite_type in neurite_types:
        for m in morphologies:
            diams = []
            lengths = []
            for neurite in m.neurites:
                cache = {}
                if neurite.type == getattr(NeuriteType, neurite_type):
                    for section in iter_sections(neurite):
                        diams.append(2 * np.mean(section.points[:, 3]))
                        tip_length = max(
                            section_path_length(_section, cache) for _section in section.ipreorder()
                        )
                        lengths.append(
                            tip_length - section_path_length(section, cache) + section.length
                        )
            lengths = np.array(lengths)
            diams = np.array(diams)
            lengths /= lengths.max()
            all_lengths[neurite_type] += lengths.tolist()
            all_diams[neurite_type] += diams.tolist()
        p, extra = Polynomial.fit(
            all_lengths[neurite_type],
            all_diams[neurite_type],
            fit_orders[neurite_type],
            full=True,
        )
        residues[neurite_type] = float(extra[0][0])
        coeffs[neurite_type] = p.convert().coef.tolist()
    return coeffs, [all_lengths, all_diams, residues]


def plot_model(coeffs, pdf, title_str, all_lengths, all_diams, residues):
    """Plot diameter model."""
    color = {"basal_dendrite": "r", "apical_dendrite": "m", "axon": "b"}
    plt.figure(figsize=(5, 4))
    for neurite_type, coeff in coeffs.items():
        xd = all_lengths[neurite_type]
        yd = all_diams[neurite_type]
        plt.scatter(xd, yd, s=0.5, marker=".", c=color[neurite_type])
        x = np.linspace(0, 1, 100)
        p = Polynomial(coeff)
        plt.plot(x, p(x), c=color[neurite_type], lw=2, label=neurite_type)
        res = np.round(residues[neurite_type], 1)
        coef = [np.round(c, 3) for c in coeff]
        pears = np.around(pearsonr(xd, yd)[0], 2)
        title_str += f"""
        {neurite_type}: fit: {coef}, residual: {res}, pearson: {pears}"""
    plt.suptitle(title_str, fontsize=5)
    plt.xlabel("normalise path distance to most distal tip")
    plt.ylabel("mean section radius")
    plt.legend()
    pdf.savefig()
    plt.close()


# pylint: disable=unused-argument
def simpler_diametrizer(morphology, neurite_types, model_params, diam_params=None, rng=np.random):
    """Diametrize a morphology."""
    if not isinstance(neurite_types, list):
        neurite_types = [neurite_types]
    neurite_types = [
        getattr(NeuriteType, neurite_type) if isinstance(neurite_type, str) else neurite_type
        for neurite_type in neurite_types
    ]
    _morphology = Morphology(morphology)
    for neurite in _morphology.neurites:
        if neurite.type in neurite_types and neurite.type.name in model_params:
            p = Polynomial(model_params[neurite.type.name])
            cache = {}
            max_len = max(terminal_path_lengths(neurite, cache))
            for section in iter_sections(neurite):
                tip_length = max(
                    section_path_length(_section, cache) for _section in section.ipreorder()
                )
                section_length = tip_length - section_path_length(section, cache) + section.length
                diam = p(section_length / max_len)
                morphology.sections[section.id].diameters = len(section.points) * [diam]
            for section in iter_sections(neurite):
                diam_start = morphology.sections[section.id].diameters[0]
                if section.children:
                    diam_end = max(
                        morphology.sections[sec.id].diameters[0] for sec in section.children
                    )
                    diam_end = 0.5 * (diam_start + diam_end)
                else:
                    diam_end = p(0)
                morphology.sections[section.id].diameters = diam_start + (
                    diam_end - diam_start
                ) * np.linspace(0, 1, len(section.points))
